/**
 * @file
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */

#include "SlaterToGaussian.h"

namespace Scine {
namespace Utils {

GtoExpansion SlaterToGaussian::getGTOExpansion(unsigned int N, unsigned int n, unsigned int l, double exponent) {
  auto values = get(N, n, l, exponent);
  GtoExpansion gtos;
  gtos.angularMomentum = l;

  gtos.gtfs.resize(N);
  for (int i = 0; i < N; i++) {
    gtos.gtfs[i] = Gtf{static_cast<int>(l), values[i].first, values[i].second};
  }

  return gtos;
}

std::vector<SlaterToGaussian::parameterPair> SlaterToGaussian::get(unsigned int N, unsigned int n, unsigned int l,
                                                                   double exponent) {
  std::vector<parameterPair> parameters(N);

  std::array<double, 6> a = {};
  std::array<double, 6> c = {};
  getValues(a, c, N, n, l);
  for (unsigned int i = 0; i < N; i++) {
    parameters[i].first = a[i] * exponent * exponent;
    parameters[i].second = c[i];
  }

  return parameters;
}

// *****************************************************************************
//>
//>  Data reproduced with permission of Prof. Dr. Jürg Hutter, 2019.
//>
// *****************************************************************************

void SlaterToGaussian::getValues(std::array<double, 6>& a, std::array<double, 6>& c, int N, int n, int l) {
  int m = (n * (n - 1)) / 2 + l + 1;
  switch (m) {
    case 1:
      sto1s(a, c, N);
      break;
    case 2:
      sto2s(a, c, N);
      break;
    case 3:
      sto2p(a, c, N);
      break;
    case 4:
      sto3s(a, c, N);
      break;
    case 5:
      sto3p(a, c, N);
      break;
    case 6:
      sto3d(a, c, N);
      break;
    case 7:
      sto4s(a, c, N);
      break;
    case 8:
      sto4p(a, c, N);
      break;
    case 9:
      sto4d(a, c, N);
      break;
    case 10:
      sto4f(a, c, N);
      break;
    case 11:
      sto5s(a, c, N);
      break;
    case 12:
      sto5p(a, c, N);
      break;
    case 13:
      sto5d(a, c, N);
      break;
    case 14:
      sto5f(a, c, N);
      break;
    case 15:
      sto5g(a, c, N);
      break;
    case 16:
      sto6s(a, c, N);
      break;
    case 17:
      sto6p(a, c, N);
      break;
    case 18:
      sto6d(a, c, N);
      break;
    case 19:
      sto6f(a, c, N);
      break;
    case 20:
      sto6g(a, c, N);
      break;
    case 21:
      sto6h(a, c, N);
      break;
    case 22:
      sto7s(a, c, N);
      break;
    case 23:
      sto7p(a, c, N);
      break;
    case 24:
      sto7d(a, c, N);
      break;
    case 25:
      sto7f(a, c, N);
      break;
    case 26:
      sto7g(a, c, N);
      break;
    case 27:
      sto7h(a, c, N);
      break;
    case 28:
      sto7i(a, c, N);
    default:;
  }
  // scaling: a = zeta^2 * a
  //  alpha (1:n) = zeta**2 * alpha (1:n)
}

// clang-format off

void SlaterToGaussian::sto7i(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.06380473723420682; c[0] = 0.9999999999999998;
      break;
    case 2:
      a[0] = 0.045457834184705205; c[0] = 0.6463784604860205;
      a[1] = 0.10559967649617007; c[1] = 0.49629977931368163;
      break;
    case 3:
      a[0] = 0.03666777792117483; c[0] = 0.3678119967003228;
      a[1] = 0.07224640498805254; c[1] = 0.618554233547922;
      a[2] = 0.15965614371413928; c[2] = 0.18501800416077063;
      break;
    case 4:
      a[0] = 0.031314605885208034; c[0] = 0.20067272567322478;
      a[1] = 0.05654362551330421; c[1] = 0.5656501384781;
      a[2] = 0.1062477818701912; c[2] = 0.3489852946890129;
      a[3] = 0.230241610829344; c[3] = 0.06217858499289526;
      break;
    case 5:
      a[0] = 0.02764326937219174; c[0] = 0.1083794411727515;
      a[1] = 0.047149043197080166; c[1] = 0.4503068342384467;
      a[2] = 0.08137277011103897; c[2] = 0.4431260931378365;
      a[3] = 0.1500423393404324; c[3] = 0.15687089041229385;
      a[4] = 0.32194304635741355; c[4] = 0.020215660073712404;
      break;
    case 6:
      a[0] = 0.024937530903194137; c[0] = 0.058696755716280226;
      a[1] = 0.04081596003413946; c[1] = 0.3330255270864295;
      a[2] = 0.06665805855772095; c[2] = 0.4640568401312629;
      a[3] = 0.11296341467059927; c[3] = 0.253753255644;
      a[4] = 0.2062966749989214; c[4] = 0.06316006418978992;
      a[5] = 0.4401320366389093; c[5] = 0.0065607991581329616;
    default:;
  }
}

void SlaterToGaussian::sto7h(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.055067144574685124; c[0] = 1.0000000000000002;
      break;
    case 2:
      a[0] = 0.039504306298949056; c[0] = 0.6244224977499361;
      a[1] = 0.08748752023471955; c[1] = 0.48947010311740946;
      break;
    case 3:
      a[0] = 0.031898703524951964; c[0] = 0.329315310335007;
      a[1] = 0.060322239914691864; c[1] = 0.6230689660925297;
      a[2] = 0.12542722211419774; c[2] = 0.1792628471936351;
      break;
    case 4:
      a[0] = 0.027133297390849163; c[0] = 0.16105237279351875;
      a[1] = 0.04709911293196238; c[1] = 0.5444503553138231;
      a[2] = 0.08417053216258696; c[2] = 0.36701928787783233;
      a[3] = 0.16769024925679932; c[3] = 0.06200372233567246;
      break;
    case 5:
      a[0] = 0.0236897013320386; c[0] = 0.07357940603188007;
      a[1] = 0.0387782492773279; c[1] = 0.39203903619828495;
      a[2] = 0.0637197820910416; c[2] = 0.46247919691910666;
      a[3] = 0.10978567271801772; c[3] = 0.1832898128285397;
      a[4] = 0.2085364358880184; c[4] = 0.023617413014962847;
      break;
    case 6:
      a[0] = 0.020820906789523693; c[0] = 0.029194538236458435;
      a[1] = 0.032494198134763036; c[1] = 0.23574088940165333;
      a[2] = 0.05023584266268183; c[2] = 0.44360291827793874;
      a[3] = 0.0794251556241889; c[3] = 0.3155567845762703;
      a[4] = 0.13148712880820534; c[4] = 0.09872391034903429;
      a[5] = 0.2359872300875659; c[5] = 0.012260115931631366;
    default:;
  }
}

void SlaterToGaussian::sto7g(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.04636088534182354; c[0] = 1.0000000000000002;
      break;
    case 2:
      a[0] = 0.033089231025105; c[0] = 0.5580805992890315;
      a[1] = 0.06689561975474233; c[1] = 0.5213501458304487;
      break;
    case 3:
      a[0] = 0.02536613332751714; c[0] = 0.1958347745511226;
      a[1] = 0.04392292490201606; c[1] = 0.6069779654592313;
      a[2] = 0.07935572931927462; c[2] = 0.28358299617643856;
      break;
    case 4:
      a[0] = 0.026003867919966516; c[0] = 0.2209616450163026;
      a[1] = 0.045349648935287534; c[1] = 0.6131566254139853;
      a[2] = 0.08207336886154683; c[2] = 0.25321528599323867;
      a[3] = 0.43630841028097767; c[3] = -0.0028051508587602996;
      break;
    case 5:
      a[0] = 0.021295076240560495; c[0] = 0.06733674728927509;
      a[1] = 0.03407297844465803; c[1] = 0.39912419436935137;
      a[2] = 0.05463343066262556; c[2] = 0.47165433606934565;
      a[3] = 0.09142718539759649; c[3] = 0.15054429419419305;
      a[4] = 0.3631735814032239; c[4] = -0.003735801628196915;
      break;
    case 6:
      a[0] = 0.02154419398101922; c[0] = 0.07294053032928545;
      a[1] = 0.03458506500828156; c[1] = 0.41255126079553106;
      a[2] = 0.055591221080949486; c[2] = 0.4632159545572935;
      a[3] = 0.0929128679746364; c[3] = 0.14023937661406255;
      a[4] = 0.3353651542417124; c[4] = -0.00381264707490568;
      a[5] = 0.8332537115987487; c[5] = -0.0002777669788215993;
    default:;
  }
}

void SlaterToGaussian::sto7f(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.037696619888011394; c[0] = 1.;
      break;
    case 2:
      a[0] = 0.02399618488506292; c[0] = 0.2938876408614713;
      a[1] = 0.0441727483070652; c[1] = 0.7461850531456674;
      break;
    case 3:
      a[0] = 0.02649072884749045; c[0] = 0.41965629582857367;
      a[1] = 0.04781482493797306; c[1] = 0.6320266162094202;
      a[2] = 0.20807007812612838; c[2] = -0.025856724476949224;
      break;
    case 4:
      a[0] = 0.019516667217357033; c[0] = 0.07760665191295739;
      a[1] = 0.031536542788218214; c[1] = 0.5138599419165162;
      a[2] = 0.0518809354792556; c[2] = 0.4658306981238731;
      a[3] = 0.18978036627517764; c[3] = -0.03032718718680631;
      break;
    case 5:
      a[0] = 0.02085272354115486; c[0] = 0.11822768394863675;
      a[1] = 0.033574888134901455; c[1] = 0.5490677063820284;
      a[2] = 0.05432263605261335; c[2] = 0.3941968220546063;
      a[3] = 0.1670441095948162; c[3] = -0.031436205691901344;
      a[4] = 0.3442595341279306; c[4] = -0.0033909812829636394;
      break;
    case 6:
      a[0] = 0.01641594422436568; c[0] = 0.01709822656839514;
      a[1] = 0.024762426761735977; c[1] = 0.22140165119709196;
      a[2] = 0.03711681387997685; c[2] = 0.5157825082425954;
      a[3] = 0.057006537132452996; c[3] = 0.30947877900909615;
      a[4] = 0.15940261392199198; c[4] = -0.032377223278152595;
      a[5] = 0.3168807735151213; c[5] = -0.004680397785899489;
    default:;
  }
}

void SlaterToGaussian::sto7d(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.02909039015051073; c[0] = 1.0000000000000002;
      break;
    case 2:
      a[0] = 0.03018482426518148; c[0] = 1.025499895914138;
      a[1] = 0.15242731708213653; c[1] = -0.0801196219224754;
      break;
    case 3:
      a[0] = 0.021405281219812042; c[0] = 0.2999132044312777;
      a[1] = 0.03568228986173447; c[1] = 0.7809338560049855;
      a[2] = 0.11574152434647746; c[2] = -0.1242352213758718;
      break;
    case 4:
      a[0] = 0.021660976213634955; c[0] = 0.3177854736353611;
      a[1] = 0.03606163962267541; c[1] = 0.7675316218690252;
      a[2] = 0.10985898319545016; c[2] = -0.11712958841599931;
      a[3] = 0.15908195611925; c[3] = -0.011747385912275846;
      break;
    case 5:
      // Warning: optimization not complete
      a[0] = 0.018501533387022245; c[0] = 0.11676114014037238;
      a[1] = 0.02923074108604399; c[1] = 0.6656817970408199;
      a[2] = 0.05423405317769654; c[2] = 600.6430437919528;
      a[3] = 0.054241824018008936; c[3] = -600.3759617830652;
      a[4] = 0.12494194517082237; c[4] = -0.09295829453419228;
      break;
    case 6:
      a[0] = 0.017457157083607015; c[0] = 0.07250557935571439;
      a[1] = 0.026593769405605233; c[1] = 0.5007763109536889;
      a[2] = 0.04008932402169768; c[2] = 0.5339957399786912;
      a[3] = 0.09560872001992007; c[3] = -0.11880032574322265;
      a[4] = 0.15425745053701811; c[4] = -0.03241452360588799;
      a[5] = 0.5964979973145561; c[5] = 0.0007012418324482315;
    default:;
  }
}

void SlaterToGaussian::sto7p(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.02056749089204262; c[0] = 1.0000000000000000;
      break;
    case 2:
      a[0] = 0.02509434622887706; c[0] = 1.1953936450048124;
      a[1] = 0.07932687651759715; c[1] = -0.33554981426999153;
      break;
    case 3:
      a[0] = 0.02216294295259616; c[0] = 0.8184285461616602;
      a[1] = 0.042045475539838104; c[1] = 0.8328277926936017;
      a[2] = 0.0622825034958455; c[2] = -0.7865628896072743;
      break;
    case 4:
      a[0] = 0.01798640692496774; c[0] = 0.2517796607641491;
      a[1] = 0.02838134507613473; c[1] = 1.0084655616939668;
      a[2] = 0.07596133909152862; c[2] = -0.40416345513837804;
      a[3] = 0.30794728288131357; c[3] = 0.011606225585276365;
      break;
    case 5:
      a[0] = 0.018253221396022562; c[0] = 0.27532738928860534;
      a[1] = 0.028708043693626412; c[1] = 0.9951443329321299;
      a[2] = 0.07061876309146031; c[2] = -0.2663034712401708;
      a[3] = 0.08392634608758946; c[3] = -0.1489548148570699;
      a[4] = 0.298415801799801; c[4] = 0.012451440490101195;
      break;
    case 6:
      // Warning: optimization not complete
      a[0] = 0.018266628106424383; c[0] = 0.2762223367567883;
      a[1] = 0.02870452146813995; c[1] = 0.992858662003434;
      a[2] = 0.07430061134905402; c[2] = -0.4029967818464786;
      a[3] = 0.18615871115244434; c[3] = -41.116221124256434;
      a[4] = 0.1861966605860832; c[4] = 41.115741037337784;
      a[5] = 0.476774875907382; c[5] = 0.0020913658772264638;
    default:;
  }
}

void SlaterToGaussian::sto7s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.012170316100569041; c[0] = 1.;

      break;
    case 2:
      a[0] = 0.02380792064880498; c[0] = 2.4733498823505897;
      a[1] = 0.0413963085083486; c[1] = -1.7500626134752972;
      break;
    case 3:
      a[0] = 0.020979847986630586; c[0] = 1.675236183601737;
      a[1] = 0.061178936229820555; c[1] = -1.147227387784115;
      a[2] = 0.1401881324437541; c[2] = 0.2121061484210291;
      break;
    case 4:
      a[0] = 0.015915368005026652; c[0] = 0.3275001969711621;
      a[1] = 0.023863997843078046; c[1] = 1.5476507021661345;
      a[2] = 0.054562166429946386; c[2] = -1.2697132490611722;
      a[3] = 0.16342646591350343; c[3] = 0.13574231777373438;
      break;
    case 5:
      a[0] = 0.01606210559622015; c[0] = 0.3489870695674181;
      a[1] = 0.024001717565201447; c[1] = 1.5335261688987925;
      a[2] = 0.05439364645164688; c[2] = -1.275299080068857;
      a[3] = 0.16456835599844608; c[3] = 0.13406783441420897;
      a[4] = 1.8632404725730192; c[4] = -0.0001818568135349486;
      break;
    case 6:
      a[0] = 0.01565231604983483; c[0] = 0.29154442199853614;
      a[1] = 0.023623292255693457; c[1] = 1.5691344169994597;
      a[2] = 0.05497538487325693; c[2] = -1.2618704056879249;
      a[3] = 0.15604786033627724; c[3] = 0.13683678315315853;
      a[4] = 0.30526699650551353; c[4] = 0.006442390493458186;
      a[5] = 0.7613613898332257; c[5] = -0.0009736214432373223;
    default:;
  }
}

void SlaterToGaussian::sto6h(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.07313314184000295; c[0] = 1.;
      break;
    case 2:
      a[0] = 0.05113303517913134; c[0] = 0.6498709211538302;
      a[1] = 0.12607272311510218; c[1] = 0.4911257514330729;
      break;
    case 3:
      a[0] = 0.040851454032819305; c[0] = 0.3746467681927208;
      a[1] = 0.08415648437016819; c[1] = 0.6132065860909666;
      a[2] = 0.1974425255793456; c[2] = 0.18179709985670622;
      break;
    case 4:
      a[0] = 0.034679493065603105; c[0] = 0.20806488992498948;
      a[1] = 0.06498507795614539; c[1] = 0.5659152898854298;
      a[2] = 0.1276799968878006; c[2] = 0.3407357791588573;
      a[3] = 0.293954336816017; c[3] = 0.061163634708810435;
      break;
    case 5:
      a[0] = 0.030487493495559674; c[0] = 0.1147257294844731;
      a[1] = 0.05371457065868773; c[1] = 0.45668261354533374;
      a[2] = 0.09619925104884211; c[2] = 0.434116528713387;
      a[3] = 0.18557752006634115; c[3] = 0.15177021621261605;
      a[4] = 0.42330005591830827; c[4] = 0.02006866296723323;
      break;
    case 6:
      a[0] = 0.027419761875036298; c[0] = 0.06355638710922437;
      a[1] = 0.04620886209790054; c[1] = 0.34336050905161936;
      a[2] = 0.07793999480997445; c[2] = 0.45863493125119364;
      a[3] = 0.13712767813357474; c[3] = 0.24456067327385683;
      a[4] = 0.2621356941062822; c[4] = 0.061018224093793164;
      a[5] = 0.5948121608710077; c[5] = 0.006618011961621055;
    default:;
  }
}

void SlaterToGaussian::sto6g(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.061536068942275354; c[0] = 1.;
      break;
    case 2:
      a[0] = 0.04332629465536664; c[0] = 0.6209033605306379;
      a[1] = 0.10013340560986961; c[1] = 0.4864065117811783;
      break;
    case 3:
      a[0] = 0.03454561337035713; c[0] = 0.3227633523004909;
      a[1] = 0.06729191664458245; c[1] = 0.6193784317147506;
      a[2] = 0.14454284915249543; c[2] = 0.1810539000206162;
      break;
    case 4:
      a[0] = 0.028983896307773004; c[0] = 0.15152805717139864;
      a[1] = 0.05135519389407211; c[1] = 0.5309701439600488;
      a[2] = 0.09369371519055944; c[2] = 0.37457949569535787;
      a[3] = 0.18939419012133174; c[3] = 0.06851383804287256;
      break;
    case 5:
      a[0] = 0.024699673908324754; c[0] = 0.06048210643335771;
      a[1] = 0.04083036114944879; c[1] = 0.3536959373883507;
      a[2] = 0.06763537320097322; c[2] = 0.46434612683985144;
      a[3] = 0.1170091091566765; c[3] = 0.2130170165201656;
      a[4] = 0.22066124899602302; c[4] = 0.0343912188428939;
      break;
    case 6:
      a[0] = 0.024944035346224053; c[0] = 0.06443914638193514;
      a[1] = 0.04138718127148023; c[1] = 0.36507904601268215;
      a[2] = 0.06886079537637349; c[2] = 0.4624056701486661;
      a[3] = 0.11972435105059721; c[3] = 0.20306165490832248;
      a[4] = 0.22639206282070937; c[4] = 0.03099725943908058;
      a[5] = 1.4892687021530484; c[5] = -0.00018343848780163463;
    default:;
  }
}

void SlaterToGaussian::sto6f(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.04999708511585302; c[0] = 1.;
      break;
    case 2:
      a[0] = 0.03443190219402774; c[0] = 0.5095525299458369;
      a[1] = 0.06993938754199006; c[1] = 0.556840322894986;
      break;
    case 3:
      a[0] = 0.022913376030207443; c[0] = 0.0844980697596352;
      a[1] = 0.040621057779302076; c[1] = 0.558907102542338;
      a[2] = 0.07551452585114922; c[2] = 0.4268730468790412;
      break;
    case 4:
      a[0] = 0.026723978066792376; c[0] = 0.18313299355518015;
      a[1] = 0.04675842108187079; c[1] = 0.5931630352672648;
      a[2] = 0.08424111895397314; c[2] = 0.29815135014219757;
      a[3] = 0.40182223710248033; c[3] = -0.009003129808147884;
      break;
    case 5:
      a[0] = 0.019820990273457767; c[0] = 0.026554164306668928;
      a[1] = 0.031943455135755784; c[1] = 0.2857152467032053;
      a[2] = 0.05180437300600356; c[2] = 0.5249457199886216;
      a[3] = 0.08839526561404966; c[3] = 0.23802584890828332;
      a[4] = 0.3861344493937384; c[4] = -0.00959390690588398;
      break;
    case 6:
      a[0] = 0.021979238503298942; c[0] = 0.05529549645466706;
      a[1] = 0.0353635379450325; c[1] = 0.36260443662220726;
      a[2] = 0.056802330945072296; c[2] = 0.47981873895702293;
      a[3] = 0.09436112630657385; c[3] = 0.17913120326169676;
      a[4] = 0.3258167395765562; c[4] = -0.010025342773471209;
      a[5] = 0.7758137452357111; c[5] = -0.0013665302206626205;
    default:;
  }
}

void SlaterToGaussian::sto6d(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.038539962557578994; c[0] = 1.0000000000000000;
      break;
    case 2:
      a[0] = 0.018343252403866257; c[0] = 0.1002697090882244;
      a[1] = 0.04095508659742878; c[1] = 0.921707450678634;
      break;
    case 3:
      a[0] = 0.026815363906634464; c[0] = 0.36430995341001843;
      a[1] = 0.048285664795074575; c[1] = 0.6902756637148024;
      a[2] = 0.20398527022566643; c[2] = -0.061643555476797514;
      break;
    case 4:
      a[0] = 0.014812592588885804; c[0] = 0.011887725165713237;
      a[1] = 0.028604392422323988; c[1] = 0.4126202273551119;
      a[2] = 0.049571759954916264; c[2] = 0.6324841000859261;
      a[3] = 0.20042444873055187; c[3] = -0.06355840930231531;
      break;
    case 5:
      a[0] = 0.02104884011402785; c[0] = 0.09247565169723151;
      a[1] = 0.03387333992146401; c[1] = 0.5183469446932626;
      a[2] = 0.05464871753445727; c[2] = 0.4583625355467117;
      a[3] = 0.16559124243612725; c[3] = -0.06557358037625106;
      a[4] = 0.3388543915705088; c[4] = -0.01073326765431463;
      break;
    case 6:
      a[0] = 0.012538741501253952; c[0] = 0.0013891559057176176;
      a[1] = 0.022411877391156923; c[1] = 0.12323543239586941;
      a[2] = 0.03502206086139318; c[2] = 0.516335137452715;
      a[3] = 0.05542414129417389; c[3] = 0.42911284700977503;
      a[4] = 0.16410056143462523; c[4] = -0.065949692922377;
      a[5] = 0.33496788346358164; c[5] = -0.01123082945384288;
    default:;
  }
}

void SlaterToGaussian::sto6p(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.02720326117498455; c[0] = 1.0000000000000000;
      break;
    case 2:
      a[0] = 0.03068554921378584; c[0] = 1.0932524470142648;
      a[1] = 0.13327144868476606; c[1] = -0.20022333756955418;
      break;
    case 3:
      a[0] = 0.021306449714000038; c[0] = 0.2589896573574047;
      a[1] = 0.03558737683798757; c[1] = 0.8948520093263641;
      a[2] = 0.1159879775604416; c[2] = -0.2541710812641519;
      break;
    case 4:
      a[0] = 0.02094853126286551; c[0] = 0.23874968368587424;
      a[1] = 0.03524009985593678; c[1] = 0.9113069782415131;
      a[2] = 0.1172381116493136; c[2] = -0.25095169849099785;
      a[3] = 1.074757975437711; c[3] = 0.0014270633206840072;
      break;
    case 5:
      a[0] = 0.02191796448535917; c[0] = 0.29958347534070895;
      a[1] = 0.036406617593622265; c[1] = 0.8669095451744081;
      a[2] = 0.10823682952141411; c[2] = -0.24403207325952317;
      a[3] = 0.1912844969312091; c[3] = -0.025610699331187835;
      a[4] = 0.605311988824089; c[4] = 0.003533893449371835;
      break;
    case 6:
      a[0] = 0.01836987035282947; c[0] = 0.0881865874022419;
      a[1] = 0.028283616813330213; c[1] = 0.58260237820929;
      a[2] = 0.0424808315746305; c[2] = 0.5465800645782257;
      a[3] = 0.08775028399997764; c[3] = -0.20802232268176662;
      a[4] = 0.14284922158351138; c[4] = -0.1117378797434228;
      a[5] = 0.6631101746841396; c[5] = 0.0028390610573703423;
    default:;
  }
}

void SlaterToGaussian::sto6s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 0.016053956888690133; c[0] = 1.000000000000000;
      break;
    case 2:
      a[0] = 0.026188993287017705; c[0] = 1.6061745480037777;
      a[1] = 0.06916086082602657; c[1] = -0.8493191121412285;
      break;
    case 3:
      a[0] = 0.025012605890022674; c[0] = 1.4570914918085252;
      a[1] = 0.08530028544682955; c[1] = -0.7792202598134607;
      a[2] = 0.23664111527549864; c[2] = 0.08944830584494995;
      break;
    case 4:
      a[0] = 0.018111879224996335; c[0] = 0.2742426110211285;
      a[1] = 0.028489853269487397; c[1] = 1.3154513257751648;
      a[2] = 0.07575858742450885; c[2] = -0.8700267823821266;
      a[3] = 0.3135211502054625; c[3] = 0.04933652686369165;
      break;
    case 5:
      a[0] = 0.017704122931955696; c[0] = 0.23979549908360312;
      a[1] = 0.02816438780280081; c[1] = 1.339995854967376;
      a[2] = 0.07635157423691516; c[2] = -0.8631396815113626;
      a[3] = 0.29378907244008734; c[3] = 0.05070584407672884;
      a[4] = 0.7609367674925651; c[4] = 0.0017452546896168622;
      break;
    case 6:
      a[0] = 0.018375086251195376; c[0] = 0.3019609612694464;
      a[1] = 0.02883213591739931; c[1] = 1.3047682665168656;
      a[2] = 0.0735171952084867; c[2] = -0.8423706115013339;
      a[3] = 0.13718508837897095; c[3] = -0.08693662797569257;
      a[4] = 0.21928060661775448; c[4] = 0.08040482815742164;
      a[5] = 0.47255402578342814; c[5] = 0.011258177324556012;
    default:;
  }
}

void SlaterToGaussian::sto5g(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 8.565417784e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 1.554531559e-1; c[0] = 4.848298074e-1;
      a[1] = 5.854079811e-2; c[1] = 6.539381621e-1;
      break;
    case 3:
      a[0] = 2.545432122e-1; c[0] = 1.780980905e-1;
      a[1] = 1.006544376e-1; c[1] = 6.063757846e-1;
      a[2] = 4.624463922e-2; c[2] = 3.828552923e-1;
      break;
    case 4:
      a[0] = 3.945205573e-1; c[0] = 6.010484250e-2;
      a[1] = 1.588100623e-1; c[1] = 3.309738329e-1;
      a[2] = 7.646521729e-2; c[2] = 5.655207585e-1;
      a[3] = 3.898703611e-2; c[3] = 2.171122608e-1;
      break;
    case 5:
      a[0] = 5.895429375e-1; c[0] = 1.998085812e-2;
      a[1] = 2.393343780e-1; c[1] = 1.460384050e-1;
      a[2] = 1.172646904e-1; c[2] = 4.230565459e-1;
      a[3] = 6.254074479e-2; c[3] = 4.635699665e-1;
      a[4] = 3.411243214e-2; c[4] = 1.226411691e-1;
      break;
    case 6:
      a[0] = 8.574668996e-1; c[0] = 6.729778096e-3;
      a[1] = 3.497184772e-1; c[1] = 5.874145170e-2;
      a[2] = 1.727917060e-1; c[2] = 2.339955227e-1;
      a[3] = 9.373643151e-2; c[3] = 4.512983737e-1;
      a[4] = 5.340032759e-2; c[4] = 3.552053926e-1;
      a[5] = 3.057364464e-2; c[5] = 6.974153145e-2;
    default:;
  }
}

void SlaterToGaussian::sto5f(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 6.952785407e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 1.156094555e-1; c[0] = 4.856637346e-1;
      a[1] = 4.778940916e-2; c[1] = 6.125980914e-1;
      break;
    case 3:
      a[0] = 1.649233885e-1; c[0] = 1.909729355e-1;
      a[1] = 7.487066646e-2; c[1] = 6.146060459e-1;
      a[2] = 3.735787219e-2; c[2] = 3.059611271e-1;
      break;
    case 4:
      a[0] = 2.017831152e-1; c[0] = 9.174268830e-2;
      a[1] = 1.001952178e-1; c[1] = 4.023496947e-1;
      a[2] = 5.447006630e-2; c[2] = 4.937432100e-1;
      a[3] = 3.037569283e-2; c[3] = 1.254001522e-1;
      break;
    case 5:
      a[0] = 1.670735676e+0; c[0] = -7.301193568e-4;
      a[1] = 2.072477219e-1; c[1] = 8.414991343e-2;
      a[2] = 1.024709357e-1; c[2] = 3.923683153e-1;
      a[3] = 5.537913898e-2; c[3] = 5.040033146e-1;
      a[4] = 3.072866652e-2; c[4] = 1.328979300e-1;
      break;
    case 6:
      a[0] = 1.334096840e+0; c[0] = -9.486751531e-4;
      a[1] = 2.372312347e-1; c[1] = 4.624275998e-2;
      a[2] = 1.269485744e-1; c[2] = 2.373699784e-1;
      a[3] = 7.290318381e-2; c[3] = 4.589112231e-1;
      a[4] = 4.351355997e-2; c[4] = 3.205010548e-1;
      a[5] = 2.598071843e-2; c[5] = 5.077063693e-2;
    default:;
  }
}

void SlaterToGaussian::sto5d(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 5.352200793e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 6.906014388e-2; c[0] = 6.539405185e-1;
      a[1] = 3.399457777e-2; c[1] = 3.948945302e-1;
      break;
    case 3:
      a[0] = 4.913352950e-1; c[0] = -2.010175008e-2;
      a[1] = 7.329090601e-2; c[1] = 5.899370608e-1;
      a[2] = 3.594209290e-2; c[2] = 4.658445960e-1;
      break;
    case 4:
      a[0] = 4.230617826e-1; c[0] = -2.421626009e-2;
      a[1] = 8.293863702e-2; c[1] = 3.937644956e-1;
      a[2] = 4.590326388e-2; c[2] = 5.489520286e-1;
      a[3] = 2.628744797e-2; c[3] = 1.190436963e-1;
      break;
    case 5:
      a[0] = 9.702946470e-1; c[0] = -3.231527611e-3;
      a[1] = 3.603270196e-1; c[1] = -2.434931372e-2;
      a[2] = 8.668717752e-2; c[2] = 3.440817054e-1;
      a[3] = 4.833708379e-2; c[3] = 5.693674376e-1;
      a[4] = 2.751899341e-2; c[4] = 1.511340183e-1;
      break;
    case 6:
      a[0] = 8.820520428e-1; c[0] = -4.097377019e-3;
      a[1] = 3.410838409e-1; c[1] = -2.508271857e-2;
      a[2] = 9.204308840e-2; c[2] = 2.648458555e-1;
      a[3] = 5.472831774e-2; c[3] = 5.097437054e-1;
      a[4] = 3.391202830e-2; c[4] = 2.654483467e-1;
      a[5] = 2.108227374e-2; c[5] = 2.623132212e-2;
    default:;
  }
}

void SlaterToGaussian::sto5p(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 3.769845216e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 2.691294191e-1; c[0] = -1.034227010e-1;
      a[1] = 3.980805011e-2; c[1] = 1.033376378e+0;
      break;
    case 3:
      a[0] = 2.127482317e-1; c[0] = -1.389529695e-1;
      a[1] = 4.729648620e-2; c[1] = 8.076691064e-1;
      a[2] = 2.604865324e-2; c[2] = 2.726029342e-1;
      break;
    case 4:
      a[0] = 3.962838833e-1; c[0] = -1.801459207e-2;
      a[1] = 1.838858552e-1; c[1] = -1.360777372e-1;
      a[2] = 4.943555157e-2; c[2] = 7.533973719e-1;
      a[3] = 2.750222273e-2; c[3] = 3.409304859e-1;
      break;
    case 5:
      a[0] = 3.422168934e-1; c[0] = -3.113958289e-2;
      a[1] = 1.665099900e-1; c[1] = -1.374007017e-1;
      a[2] = 5.443732013e-2; c[2] = 5.573881018e-1;
      a[3] = 3.367775277e-2; c[3] = 4.855428100e-1;
      a[4] = 2.091949042e-2; c[4] = 6.605423564e-2;
      break;
    case 6:
      a[0] = 3.778623374e+0; c[0] = 1.163246387e-4;
      a[1] = 3.499121109e-1; c[1] = -2.920771322e-2;
      a[2] = 1.683175469e-1; c[2] = -1.381051233e-1;
      a[3] = 5.404070736e-2; c[3] = 5.706134877e-1;
      a[4] = 3.328911801e-2; c[4] = 4.768808140e-1;
      a[5] = 2.063815019e-2; c[5] = 6.021665516e-2;
    default:;
  }
}

void SlaterToGaussian::sto5s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 2.216912938e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 1.213425654e-1; c[0] = -5.114756049e-1;
      a[1] = 3.133152144e-2; c[1] = 1.307377277e+0;
      break;
    case 3:
      a[0] = 1.080198458e-1; c[0] = -6.617401158e-1;
      a[1] = 4.408119328e-2; c[1] = 7.467595004e-1;
      a[2] = 2.610811810e-2; c[2] = 7.146490945e-1;
      break;
    case 4:
      a[0] = 8.602284252e-1; c[0] = 1.103657561e-2;
      a[1] = 1.189050200e-1; c[1] = -5.606519023e-1;
      a[2] = 3.446076176e-2; c[2] = 1.179429987e+0;
      a[3] = 1.974798796e-2; c[3] = 1.734974376e-1;
      break;
    case 5:
      a[0] = 7.403763257e-1; c[0] = 1.375523371e-2;
      a[1] = 1.367990863e-1; c[1] = -3.097344179e-1;
      a[2] = 9.135301779e-2; c[2] = -3.199192259e-1;
      a[3] = 3.726907315e-2; c[3] = 1.084547038e+0;
      a[4] = 2.241490836e-2; c[4] = 3.345288361e-1;
      break;
    case 6:
      a[0] = 1.410128298e+0; c[0] = 2.695439582e-3;
      a[1] = 5.077878915e-1; c[1] = 1.850157487e-2;
      a[2] = 1.847926858e-1; c[2] = -9.588628125e-2;
      a[3] = 1.061070594e-1; c[3] = -5.200673560e-1;
      a[4] = 3.669584901e-2; c[4] = 1.087619490e+0;
      a[5] = 2.213558430e-2; c[5] = 3.103964343e-1;
    default:;
  }
}

void SlaterToGaussian::sto4f(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 1.033434062e-1; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 2.006693538e-1; c[0] = 4.769346276e-1;
      a[1] = 6.865384900e-2; c[1] = 6.587383976e-1;
      break;
    case 3:
      a[0] = 3.483826963e-1; c[0] = 1.737856685e-1;
      a[1] = 1.249380537e-1; c[1] = 5.973380628e-1;
      a[2] = 5.349995725e-2; c[2] = 3.929395614e-1;
      break;
    case 4:
      a[0] = 5.691670217e-1; c[0] = 5.902730589e-2;
      a[1] = 2.074585819e-1; c[1] = 3.191828952e-1;
      a[2] = 9.298346885e-2; c[2] = 5.639423893e-1;
      a[3] = 4.473508853e-2; c[3] = 2.284796537e-1;
      break;
    case 5:
      a[0] = 8.925960415e-1; c[0] = 1.999839052e-2;
      a[1] = 3.277589120e-1; c[1] = 1.395427440e-1;
      a[2] = 1.492869962e-1; c[2] = 4.091508237e-1;
      a[3] = 7.506099109e-2; c[3] = 4.708252119e-1;
      a[4] = 3.892475795e-2; c[4] = 1.328082566e-1;
      break;
    case 6:
      a[0] = 1.357718039e+0; c[0] = 6.930234381e-3;
      a[1] = 5.004907278e-1; c[1] = 5.634263745e-2;
      a[2] = 2.296565064e-1; c[2] = 2.217065797e-1;
      a[3] = 1.173146814e-1; c[3] = 4.411388883e-1;
      a[4] = 6.350097171e-2; c[4] = 3.688112625e-1;
      a[5] = 3.474556673e-2; c[5] = 7.787514504e-2;
    default:;
  }
}

void SlaterToGaussian::sto4d(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 7.941656339e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 1.330958892e-1; c[0] = 4.932764167e-1;
      a[1] = 5.272119659e-2; c[1] = 5.918727866e-1;
      break;
    case 3:
      a[0] = 1.777717219e-1; c[0] = 2.308552718e-1;
      a[1] = 8.040647350e-2; c[1] = 6.042409177e-1;
      a[2] = 3.949855551e-2; c[2] = 2.595768926e-1;
      break;
    case 4:
      a[0] = 1.995825422e+0; c[0] = -2.816702620e-3;
      a[1] = 1.823461280e-1; c[1] = 2.177095871e-1;
      a[2] = 8.197240896e-2; c[2] = 6.058047348e-1;
      a[3] = 4.000634951e-2; c[3] = 2.717811257e-1;
      break;
    case 5:
      a[0] = 1.522122079e+0; c[0] = -3.673711876e-3;
      a[1] = 2.173041823e-1; c[1] = 1.167122499e-1;
      a[2] = 1.084876577e-1; c[2] = 4.216476416e-1;
      a[3] = 5.836797641e-2; c[3] = 4.547673415e-1;
      a[4] = 3.206682246e-2; c[4] = 1.037803318e-1;
      break;
    case 6:
      a[0] = 4.634239420e+0; c[0] = -4.749842876e-4;
      a[1] = 1.341648295e+0; c[1] = -3.566777891e-3;
      a[2] = 2.209593028e-1; c[2] = 1.108670481e-1;
      a[3] = 1.101467943e-1; c[3] = 4.159646930e-1;
      a[4] = 5.904190370e-2; c[4] = 4.621672517e-1;
      a[5] = 3.232628887e-2; c[5] = 1.081250196e-1;
    default:;
  }
}

void SlaterToGaussian::sto4p(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 5.578350235e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 6.190052680e-2; c[0] = 8.743116767e-1;
      a[1] = 2.648418407e-2; c[1] = 1.513640107e-1;
      break;
    case 3:
      a[0] = 4.859692220e-1; c[0] = -6.147823411e-2;
      a[1] = 7.430216918e-2; c[1] = 6.604172234e-1;
      a[2] = 3.653340923e-2; c[2] = 3.932639495e-1;
      break;
    case 4:
      a[0] = 1.492607880e+0; c[0] = -6.035216774e-3;
      a[1] = 4.327619272e-1; c[1] = -6.013310874e-2;
      a[2] = 7.553156064e-2; c[2] = 6.451518200e-1;
      a[3] = 3.706272183e-2; c[3] = 4.117923820e-1;
      break;
    case 5:
      a[0] = 1.091977298e+0; c[0] = -1.143929558e-2;
      a[1] = 3.719985051e-1; c[1] = -6.322651538e-2;
      a[2] = 8.590019352e-2; c[2] = 4.398907721e-1;
      a[3] = 4.786503860e-2; c[3] = 5.245859166e-1;
      a[4] = 2.730479990e-2; c[4] = 1.017072253e-1;
      break;
    case 6:
      a[0] = 2.389722618e+0; c[0] = -1.665913575e-3;
      a[1] = 7.960947826e-1; c[1] = -1.657464971e-2;
      a[2] = 3.415541380e-1; c[2] = -5.958513378e-2;
      a[3] = 8.847434525e-2; c[3] = 4.053115554e-1;
      a[4] = 4.958248334e-2; c[4] = 5.433958189e-1;
      a[5] = 2.816929784e-2; c[5] = 1.204970491e-1;
    default:;
  }
}

void SlaterToGaussian::sto4s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 3.264600274e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 2.441785453e-1; c[0] = -3.046656896e-1;
      a[1] = 4.051097664e-2; c[1] = 1.146877294e+0;
      break;
    case 3:
      a[0] = 2.267938753e-1; c[0] = -3.349048323e-1;
      a[1] = 4.448178019e-2; c[1] = 1.056744667e+0;
      a[2] = 2.195294664e-2; c[2] = 1.256661680e-1;
      break;
    case 4:
      a[0] = 3.242212833e-1; c[0] = -1.120682822e-1;
      a[1] = 1.663217177e-1; c[1] = -2.845426863e-1;
      a[2] = 5.081097451e-2; c[2] = 8.909873788e-1;
      a[3] = 2.829066600e-2; c[3] = 3.517811205e-1;
      break;
    case 5:
      a[0] = 2.980263783e+0; c[0] = 1.513948997e-3;
      a[1] = 3.792228833e-1; c[1] = -7.316801548e-2;
      a[2] = 1.789717224e-1; c[2] = -3.143703799e-1;
      a[3] = 5.002110360e-2; c[3] = 9.032615169e-1;
      a[4] = 2.789361681e-2; c[4] = 3.294210848e-1;
      break;
    case 6:
      a[0] = 3.232838646e+0; c[0] = 1.374817488e-3;
      a[1] = 3.605788802e-1; c[1] = -8.666390043e-2;
      a[2] = 1.717905487e-1; c[2] = -3.130627309e-1;
      a[3] = 5.277666487e-2; c[3] = 7.812787397e-1;
      a[4] = 3.163400284e-2; c[4] = 4.389247988e-1;
      a[5] = 1.874093091e-2; c[5] = 2.487178756e-2;
    default:;
  }
}

void SlaterToGaussian::sto3d(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 1.302270363e-1; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 2.777427345e-1; c[0] = 4.666137923e-1;
      a[1] = 8.336507714e-2; c[1] = 6.644706516e-1;
      break;
    case 3:
      a[0] = 5.229112225e-1; c[0] = 1.696596060e-1;
      a[1] = 1.639595876e-1; c[1] = 5.847984817e-1;
      a[2] = 6.386630021e-2; c[2] = 4.056779523e-1;
      break;
    case 4:
      a[0] = 9.185846715e-1; c[0] = 5.799057705e-2;
      a[1] = 2.920461109e-1; c[1] = 3.045581349e-1;
      a[2] = 1.187568890e-1; c[2] = 5.601358038e-1;
      a[3] = 5.286755896e-2; c[3] = 2.432423313e-1;
      break;
    case 5:
      a[0] = 1.539033958e+0; c[0] = 2.020869128e-2;
      a[1] = 4.922090297e-1; c[1] = 1.321157923e-1;
      a[2] = 2.029756928e-1; c[2] = 3.911240346e-1;
      a[3] = 9.424112917e-2; c[3] = 4.779609701e-1;
      a[4] = 4.569058269e-2; c[4] = 1.463662294e-1;
      break;
    case 6:
      a[0] = 2.488296923e+0; c[0] = 7.283828112e-3;
      a[1] = 7.981487853e-1; c[1] = 5.386799363e-2;
      a[2] = 3.311327490e-1; c[2] = 2.072139149e-1;
      a[3] = 1.559114463e-1; c[3] = 4.266269092e-1;
      a[4] = 7.877734732e-2; c[4] = 3.843100204e-1;
      a[5] = 4.058484363e-2; c[5] = 8.902827546e-2;
    default:;
  }
}

void SlaterToGaussian::sto3p(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 9.113614253e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 1.458620964e-1; c[0] = 5.349653144e-1;
      a[1] = 5.664210742e-2; c[1] = 5.299607212e-1;
      break;
    case 3:
      a[0] = 2.692880368e+0; c[0] = -1.061945788e-2;
      a[1] = 1.489359592e-1; c[1] = 5.218564264e-1;
      a[2] = 5.739585040e-2; c[2] = 5.450015143e-1;
      break;
    case 4:
      a[0] = 1.853180239e+0; c[0] = -1.434249391e-2;
      a[1] = 1.915075719e-1; c[1] = 2.755177589e-1;
      a[2] = 8.655487938e-2; c[2] = 5.846750879e-1;
      a[3] = 4.184253862e-2; c[3] = 2.144986514e-1;
      break;
    case 5:
      a[0] = 6.466803859e+0; c[0] = -2.329023747e-3;
      a[1] = 1.555914802e+0; c[1] = -1.357395221e-2;
      a[2] = 1.955925255e-1; c[2] = 2.632185383e-1;
      a[3] = 8.809647701e-2; c[3] = 5.880427024e-1;
      a[4] = 4.234835707e-2; c[4] = 2.242794445e-1;
      break;
    case 6:
      a[0] = 5.077973607e+0; c[0] = -3.329929840e-3;
      a[1] = 1.340786940e+0; c[1] = -1.419488340e-2;
      a[2] = 2.248434849e-1; c[2] = 1.639395770e-1;
      a[3] = 1.131741848e-1; c[3] = 4.485358256e-1;
      a[4] = 6.076408893e-2; c[4] = 3.908813050e-1;
      a[5] = 3.315424265e-2; c[5] = 7.411456232e-2;
    default:;
  }
}

void SlaterToGaussian::sto3s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 5.296881757e-2; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 6.694095822e-1; c[0] = -1.529645716e-1;
      a[1] = 5.837135094e-2; c[1] = 1.051370110e+0;
      break;
    case 3:
      a[0] = 5.641487709e-1; c[0] = -1.782577972e-1;
      a[1] = 6.924421391e-2; c[1] = 8.612761663e-1;
      a[2] = 3.269529097e-2; c[2] = 2.261841969e-1;
      break;
    case 4:
      a[0] = 1.513265591e+0; c[0] = -3.295496352e-2;
      a[1] = 4.262497508e-1; c[1] = -1.724516959e-1;
      a[2] = 7.643320863e-2; c[2] = 7.518511194e-1;
      a[3] = 3.760545063e-2; c[3] = 3.589627310e-1;
      break;
    case 5:
      a[0] = 4.275877914e+0; c[0] = -3.920358850e-3;
      a[1] = 1.132409433e+0; c[1] = -4.168430506e-2;
      a[2] = 4.016256968e-1; c[2] = -1.637440990e-1;
      a[3] = 7.732370620e-2; c[3] = 7.419373723e-1;
      a[4] = 3.800708627e-2; c[4] = 3.724364929e-1;
      break;
    case 6:
      a[0] = 3.273031938e+0; c[0] = -6.775596947e-3;
      a[1] = 9.200611311e-1; c[1] = -5.639325779e-2;
      a[2] = 3.593349765e-1; c[2] = -1.587856086e-1;
      a[3] = 8.636686991e-2; c[3] = 5.534527651e-1;
      a[4] = 4.797373812e-2; c[4] = 5.015351020e-1;
      a[5] = 2.724741144e-2; c[5] = 7.223633674e-2;
    default:;
  }
}

void SlaterToGaussian::sto2p(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 1.759666885e-1; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 4.323908358e-1; c[0] = 4.522627513e-1;
      a[1] = 1.069439065e-1; c[1] = 6.713122642e-1;
      break;
    case 3:
      a[0] = 9.192379002e-1; c[0] = 1.623948553e-1;
      a[1] = 2.359194503e-1; c[1] = 5.661708862e-1;
      a[2] = 8.009805746e-2; c[2] = 4.223071752e-1;
      break;
    case 4:
      a[0] = 1.798260992e+0; c[0] = 5.713170255e-2;
      a[1] = 4.662622228e-1; c[1] = 2.857455515e-1;
      a[2] = 1.643718620e-1; c[2] = 5.517873105e-1;
      a[3] = 6.543927065e-2; c[3] = 2.632314924e-1;
      break;
    case 5:
      a[0] = 3.320386533e+0; c[0] = 2.079051117e-2;
      a[1] = 8.643257633e-1; c[1] = 1.235472099e-1;
      a[2] = 3.079819284e-1; c[2] = 3.667738886e-1;
      a[3] = 1.273309895e-1; c[3] = 4.834930290e-1;
      a[4] = 5.606243164e-2; c[4] = 1.653444074e-1;
      break;
    case 6:
      a[0] = 5.868285913e+0; c[0] = 7.924233646e-3;
      a[1] = 1.530329631e+0; c[1] = 5.144104825e-2;
      a[2] = 5.475665231e-1; c[2] = 1.898400060e-1;
      a[3] = 2.288932733e-1; c[3] = 4.049863191e-1;
      a[4] = 1.046655969e-1; c[4] = 4.012362861e-1;
      a[5] = 4.948220127e-2; c[5] = 1.051855189e-1;
    default:;
  }
}

void SlaterToGaussian::sto2s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 1.012151084e-1; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 1.292278611e-1; c[0] = 7.470867124e-1;
      a[1] = 4.908584205e-2; c[1] = 2.855980556e-1;
      break;
    case 3:
      a[0] = 2.581578398e+0; c[0] = -5.994474934e-2;
      a[1] = 1.567622104e-1; c[1] = 5.960385398e-1;
      a[2] = 6.018332272e-2; c[2] = 4.581786291e-1;
      break;
    case 4:
      a[0] = 1.161525551e+1; c[0] = -1.198411747e-2;
      a[1] = 2.000243111e+0; c[1] = -5.472052539e-2;
      a[2] = 1.607280687e-1; c[2] = 5.805587176e-1;
      a[3] = 6.125744532e-2; c[3] = 4.770079976e-1;
      break;
    case 5:
      a[0] = 8.984956862e+0; c[0] = -1.596349096e-2;
      a[1] = 1.673710636e+0; c[1] = -5.685884883e-2;
      a[2] = 1.944726668e-1; c[2] = 3.698265599e-1;
      a[3] = 8.806345634e-2; c[3] = 5.480512593e-1;
      a[4] = 4.249068522e-2; c[4] = 1.472634893e-1;
      break;
    case 6:
      a[0] = 2.768496241e+1; c[0] = -4.151277819e-3;
      a[1] = 5.077140627e+0; c[1] = -2.067024148e-2;
      a[2] = 1.426786050e+0; c[2] = -5.150303337e-2;
      a[3] = 2.040335729e-1; c[3] = 3.346271174e-1;
      a[4] = 9.260298399e-2; c[4] = 5.621061301e-1;
      a[5] = 4.416183978e-2; c[5] = 1.712994697e-1;
    default:;
  }
}

void SlaterToGaussian::sto1s(std::array<double, 6>& a, std::array<double, 6>& c, int N) {
  switch (N) {
    case 1:
      a[0] = 2.709496091e-1; c[0] = 1.000000000e+0;
      break;
    case 2:
      a[0] = 8.518186635e-1; c[0] = 4.301284983e-1;
      a[1] = 1.516232927e-1; c[1] = 6.789135305e-1;
      break;
    case 3:
      a[0] = 2.227660584e+0; c[0] = 1.543289673e-1;
      a[1] = 4.057711562e-1; c[1] = 5.353281423e-1;
      a[2] = 1.098175104e-1; c[2] = 4.446345422e-1;
      break;
    case 4:
      a[0] = 5.216844534e+0; c[0] = 5.675242080e-2;
      a[1] = 9.546182760e-1; c[1] = 2.601413550e-1;
      a[2] = 2.652034102e-1; c[2] = 5.328461143e-1;
      a[3] = 8.801862774e-2; c[3] = 2.916254405e-1;
      break;
    case 5:
      a[0] = 1.130563696e+1; c[0] = 2.214055312e-2;
      a[1] = 2.071728178e+0; c[1] = 1.135411520e-1;
      a[2] = 5.786484833e-1; c[2] = 3.318161484e-1;
      a[3] = 1.975724573e-1; c[3] = 4.825700713e-1;
      a[4] = 7.445271746e-2; c[4] = 1.935721966e-1;
      break;
    case 6:
      a[0] = 2.310303149e+1; c[0] = 9.163596280e-3;
      a[1] = 4.235915534e+0; c[1] = 4.936149294e-2;
      a[2] = 1.185056519e+0; c[2] = 1.685383049e-1;
      a[3] = 4.070988982e-1; c[3] = 3.705627997e-1;
      a[4] = 1.580884151e-1; c[4] = 4.164915298e-1;
      a[5] = 6.510953954e-2; c[5] = 1.303340841e-1;
    default:;
  }
}

// clang-format on

} // namespace Utils
} // namespace Scine
