/**
 * @file ElementTypes.h
 * @copyright This code is licensed under the 3-clause BSD license.\n
 *            Copyright ETH Zurich, Laboratory of Physical Chemistry, Reiher Group.\n
 *            See LICENSE.txt for details.
 */
#ifndef UTILS_ELEMENTTYPES_H_
#define UTILS_ELEMENTTYPES_H_

namespace Scine {
namespace Utils {

namespace detail {

// For isotopes, put the atomic mass number in the enum's bit representation
constexpr unsigned isotope(const unsigned Z, const unsigned A) noexcept {
  // Z ranges from up to 112 < 128 = 2^7
  constexpr unsigned zBits = 7;
  // A ranges up to 294 < 512 = 2^9
  // so we need at least 16 bits to represent both in the underlying value
  static_assert(sizeof(unsigned) >= 2, "Unsigned type is smaller than two bytes, isotope element types doesn't work");

  return Z + (A << zBits);
}

} // namespace detail

// clang-format off
/**
 * @brief Enum class defining all elements and isotopes
 */
enum class ElementType : unsigned {
  none = 0,
  // 1s
  H = 1,
  He = 2,
  // 2s
  Li = 3,
  Be = detail::isotope(4, 9),
  // 2p
  B = 5,
  C = 6,
  N = 7,
  O = 8,
  F = detail::isotope(9, 19),
  Ne = 10,
  // 3s
  Na = detail::isotope(11, 23),
  Mg = 12,
  // 3p
  Al = detail::isotope(13, 27),
  Si = 14,
  P = detail::isotope(15, 31),
  S = 16,
  Cl = 17,
  Ar = 18,
  // 4s
  K = 19,
  Ca = 20,
  // 3d
  Sc = detail::isotope(21, 45),
  Ti = 22,
  V = 23,
  Cr = 24,
  Mn = detail::isotope(25, 55),
  Fe = 26,
  Co = detail::isotope(27, 59),
  Ni = 28,
  Cu = 29,
  Zn = 30,
  // 4p
  Ga = 31,
  Ge = 32,
  As = detail::isotope(33, 75),
  Se = 34,
  Br = 35,
  Kr = 36,
  // 5s
  Rb = 37,
  Sr = 38,
  // 4d
  Y = detail::isotope(39, 89),
  Zr = 40,
  Nb = detail::isotope(41, 93),
  Mo = 42,
  Tc = 43,
  Ru = 44,
  Rh = detail::isotope(45, 103),
  Pd = 46,
  Ag = 47,
  Cd = 48,
  // 5p
  In = 49,
  Sn = 50,
  Sb = 51,
  Te = 52,
  I = detail::isotope(53, 127),
  Xe = 54,
  // 6s
  Cs = detail::isotope(55, 133),
  Ba = 56,
  // 4f
  La = 57,
  Ce = 58,
  Pr = detail::isotope(59, 141),
  Nd = 60,
  Pm = 61,
  Sm = 62,
  Eu = 63,
  Gd = 64,
  Tb = detail::isotope(65, 159),
  Dy = 66,
  Ho = detail::isotope(67, 165),
  Er = 68,
  Tm = detail::isotope(69, 169),
  Yb = 70,
  // 5d
  Lu = 71,
  Hf = 72,
  Ta = 73,
  W  = 74,
  Re = 75,
  Os = 76,
  Ir = 77,
  Pt = 78,
  Au = detail::isotope(79, 197),
  Hg = 80,
  // 6p
  Tl = 81,
  Pb = 82,
  Bi = detail::isotope(83, 209),
  Po = 84,
  At = 85,
  Rn = 86,
  // 7s
  Fr = detail::isotope(87, 223),
  Ra = 88,
  // 5f
  Ac = detail::isotope(89, 227),
  Th = 90,
  Pa = detail::isotope(91, 231),
   U = 92,
  Np = 93,
  Pu = 94,
  Am = 95,
  Cm = 96,
  Bk = 97,
  Cf = 98,
  Es = detail::isotope(99, 252),
  Fm = detail::isotope(100, 257),
  Md = 101,
  No = detail::isotope(102, 259),
  // 6d
  Lr = detail::isotope(103, 262),
  Rf = detail::isotope(104, 267),
  Db = detail::isotope(105, 268),
  Sg = detail::isotope(106, 271),
  Bh = detail::isotope(107, 272),
  Hs = detail::isotope(108, 270),
  Mt = detail::isotope(109, 276),
  Ds = detail::isotope(110, 281),
  Rg = detail::isotope(111, 280),
  Cn = detail::isotope(112, 285),


  // Isotopes
  H1 = detail::isotope(1, 1),
  D = detail::isotope(1, 2),
  T = detail::isotope(1, 3),
  He3 = detail::isotope(2, 3),
  He4 = detail::isotope(2, 4),
  Li6 = detail::isotope(3, 6),
  Li7 = detail::isotope(3, 7),
  Be9 = Be,
  B10 = detail::isotope(5, 10),
  B11 = detail::isotope(5, 11),
  C12 = detail::isotope(6, 12),
  C13 = detail::isotope(6, 13),
  C14 = detail::isotope(6, 14),
  N14 = detail::isotope(7, 14),
  N15 = detail::isotope(7, 15),
  O16 = detail::isotope(8, 16),
  O17 = detail::isotope(8, 17),
  O18 = detail::isotope(8, 18),
  F19 = F,
  Ne20 = detail::isotope(10, 20),
  Ne21 = detail::isotope(10, 21),
  Ne22 = detail::isotope(10, 22),
  Na23 = Na,
  Mg24 = detail::isotope(12, 24),
  Mg25 = detail::isotope(12, 25),
  Mg26 = detail::isotope(12, 26),
  Al27 = Al,
  Si28 = detail::isotope(14, 28),
  Si29 = detail::isotope(14, 29),
  Si30 = detail::isotope(14, 30),
  P31 = P,
  S32 = detail::isotope(16, 32),
  S33 = detail::isotope(16, 33),
  S34 = detail::isotope(16, 34),
  S36 = detail::isotope(16, 36),
  Cl35 = detail::isotope(17, 35),
  Cl37 = detail::isotope(17, 37),
  Ar36 = detail::isotope(18, 36),
  Ar38 = detail::isotope(18, 38),
  Ar40 = detail::isotope(18, 40),
  K39 = detail::isotope(19, 39),
  K40 = detail::isotope(19, 40),
  K41 = detail::isotope(19, 41),
  Ca40 = detail::isotope(20, 40),
  Ca42 = detail::isotope(20, 42),
  Ca43 = detail::isotope(20, 43),
  Ca44 = detail::isotope(20, 44),
  Ca46 = detail::isotope(20, 46),
  Ca48 = detail::isotope(20, 48),
  Sc45 = Sc,
  Ti46 = detail::isotope(22, 46),
  Ti47 = detail::isotope(22, 47),
  Ti48 = detail::isotope(22, 48),
  Ti49 = detail::isotope(22, 49),
  Ti50 = detail::isotope(22, 50),
  V50 = detail::isotope(23, 50),
  V51 = detail::isotope(23, 51),
  Cr50 = detail::isotope(24, 50),
  Cr52 = detail::isotope(24, 52),
  Cr53 = detail::isotope(24, 53),
  Cr54 = detail::isotope(24, 54),
  Mn55 = Mn,
  Fe54 = detail::isotope(26, 54),
  Fe56 = detail::isotope(26, 56),
  Fe57 = detail::isotope(26, 57),
  Fe58 = detail::isotope(26, 58),
  Co59 = Co,
  Ni58 = detail::isotope(28, 58),
  Ni60 = detail::isotope(28, 60),
  Ni61 = detail::isotope(28, 61),
  Ni62 = detail::isotope(28, 62),
  Ni64 = detail::isotope(28, 64),
  Cu63 = detail::isotope(29, 63),
  Cu65 = detail::isotope(29, 65),
  Zn64 = detail::isotope(30, 64),
  Zn66 = detail::isotope(30, 66),
  Zn67 = detail::isotope(30, 67),
  Zn68 = detail::isotope(30, 68),
  Zn70 = detail::isotope(30, 70),
  Ga69 = detail::isotope(31, 69),
  Ga71 = detail::isotope(31, 71),
  Ge70 = detail::isotope(32, 70),
  Ge72 = detail::isotope(32, 72),
  Ge73 = detail::isotope(32, 73),
  Ge74 = detail::isotope(32, 74),
  Ge76 = detail::isotope(32, 76),
  As75 = As,
  Se74 = detail::isotope(34, 74),
  Se76 = detail::isotope(34, 76),
  Se77 = detail::isotope(34, 77),
  Se78 = detail::isotope(34, 78),
  Se80 = detail::isotope(34, 80),
  Se82 = detail::isotope(34, 82),
  Br79 = detail::isotope(35, 79),
  Br81 = detail::isotope(35, 81),
  Kr78 = detail::isotope(36, 78),
  Kr80 = detail::isotope(36, 80),
  Kr82 = detail::isotope(36, 82),
  Kr83 = detail::isotope(36, 83),
  Kr84 = detail::isotope(36, 84),
  Kr86 = detail::isotope(36, 86),
  Rb85 = detail::isotope(37, 85),
  Rb87 = detail::isotope(37, 87),
  Sr84 = detail::isotope(38, 84),
  Sr86 = detail::isotope(38, 86),
  Sr87 = detail::isotope(38, 87),
  Sr88 = detail::isotope(38, 88),
  Y89 = Y,
  Zr90 = detail::isotope(40, 90),
  Zr91 = detail::isotope(40, 91),
  Zr92 = detail::isotope(40, 92),
  Zr94 = detail::isotope(40, 94),
  Zr96 = detail::isotope(40, 96),
  Nb93 = Nb,
  Mo92 = detail::isotope(42, 92),
  Mo94 = detail::isotope(42, 94),
  Mo95 = detail::isotope(42, 95),
  Mo96 = detail::isotope(42, 96),
  Mo97 = detail::isotope(42, 97),
  Mo98 = detail::isotope(42, 98),
  Mo100 = detail::isotope(42, 100),
  Tc97 = detail::isotope(43, 97),
  Tc98 = detail::isotope(43, 98),
  Tc99 = detail::isotope(43, 99),
  Ru96 = detail::isotope(44, 96),
  Ru98 = detail::isotope(44, 98),
  Ru99 = detail::isotope(44, 99),
  Ru100 = detail::isotope(44, 100),
  Ru101 = detail::isotope(44, 101),
  Ru102 = detail::isotope(44, 102),
  Ru104 = detail::isotope(44, 104),
  Rh103 = Rh,
  Pd102 = detail::isotope(46, 102),
  Pd104 = detail::isotope(46, 104),
  Pd105 = detail::isotope(46, 105),
  Pd106 = detail::isotope(46, 106),
  Pd108 = detail::isotope(46, 108),
  Pd110 = detail::isotope(46, 110),
  Ag107 = detail::isotope(47, 107),
  Ag109 = detail::isotope(47, 109),
  Cd106 = detail::isotope(48, 106),
  Cd108 = detail::isotope(48, 108),
  Cd110 = detail::isotope(48, 110),
  Cd111 = detail::isotope(48, 111),
  Cd112 = detail::isotope(48, 112),
  Cd113 = detail::isotope(48, 113),
  Cd114 = detail::isotope(48, 114),
  Cd116 = detail::isotope(48, 116),
  In113 = detail::isotope(49, 113),
  In115 = detail::isotope(49, 115),
  Sn112 = detail::isotope(50, 112),
  Sn114 = detail::isotope(50, 114),
  Sn115 = detail::isotope(50, 115),
  Sn116 = detail::isotope(50, 116),
  Sn117 = detail::isotope(50, 117),
  Sn118 = detail::isotope(50, 118),
  Sn119 = detail::isotope(50, 119),
  Sn120 = detail::isotope(50, 120),
  Sn122 = detail::isotope(50, 122),
  Sn124 = detail::isotope(50, 124),
  Sb121 = detail::isotope(51, 121),
  Sb123 = detail::isotope(51, 123),
  Te120 = detail::isotope(52, 120),
  Te122 = detail::isotope(52, 122),
  Te123 = detail::isotope(52, 123),
  Te124 = detail::isotope(52, 124),
  Te125 = detail::isotope(52, 125),
  Te126 = detail::isotope(52, 126),
  Te128 = detail::isotope(52, 128),
  Te130 = detail::isotope(52, 130),
  I127 = I,
  Xe124 = detail::isotope(54, 124),
  Xe126 = detail::isotope(54, 126),
  Xe128 = detail::isotope(54, 128),
  Xe129 = detail::isotope(54, 129),
  Xe130 = detail::isotope(54, 130),
  Xe131 = detail::isotope(54, 131),
  Xe132 = detail::isotope(54, 132),
  Xe134 = detail::isotope(54, 134),
  Xe136 = detail::isotope(54, 136),
  Cs133 = Cs,
  Ba130 = detail::isotope(56, 130),
  Ba132 = detail::isotope(56, 132),
  Ba134 = detail::isotope(56, 134),
  Ba135 = detail::isotope(56, 135),
  Ba136 = detail::isotope(56, 136),
  Ba137 = detail::isotope(56, 137),
  Ba138 = detail::isotope(56, 138),
  La138 = detail::isotope(57, 138),
  La139 = detail::isotope(57, 139),
  Ce136 = detail::isotope(58, 136),
  Ce138 = detail::isotope(58, 138),
  Ce140 = detail::isotope(58, 140),
  Ce142 = detail::isotope(58, 142),
  Pr141 = Pr,
  Nd142 = detail::isotope(60, 142),
  Nd143 = detail::isotope(60, 143),
  Nd144 = detail::isotope(60, 144),
  Nd145 = detail::isotope(60, 145),
  Nd146 = detail::isotope(60, 146),
  Nd148 = detail::isotope(60, 148),
  Nd150 = detail::isotope(60, 150),
  Pm145 = detail::isotope(61, 145),
  Pm147 = detail::isotope(61, 147),
  Sm144 = detail::isotope(62, 144),
  Sm147 = detail::isotope(62, 147),
  Sm148 = detail::isotope(62, 148),
  Sm149 = detail::isotope(62, 149),
  Sm150 = detail::isotope(62, 150),
  Sm152 = detail::isotope(62, 152),
  Sm154 = detail::isotope(62, 154),
  Eu151 = detail::isotope(63, 151),
  Eu153 = detail::isotope(63, 153),
  Gd152 = detail::isotope(64, 152),
  Gd154 = detail::isotope(64, 154),
  Gd155 = detail::isotope(64, 155),
  Gd156 = detail::isotope(64, 156),
  Gd157 = detail::isotope(64, 157),
  Gd158 = detail::isotope(64, 158),
  Gd160 = detail::isotope(64, 160),
  Tb159 = Tb,
  Dy156 = detail::isotope(66, 156),
  Dy158 = detail::isotope(66, 158),
  Dy160 = detail::isotope(66, 160),
  Dy161 = detail::isotope(66, 161),
  Dy162 = detail::isotope(66, 162),
  Dy163 = detail::isotope(66, 163),
  Dy164 = detail::isotope(66, 164),
  Ho165 = Ho,
  Er162 = detail::isotope(68, 162),
  Er164 = detail::isotope(68, 164),
  Er166 = detail::isotope(68, 166),
  Er167 = detail::isotope(68, 167),
  Er168 = detail::isotope(68, 168),
  Er170 = detail::isotope(68, 170),
  Tm169 = Tm,
  Yb168 = detail::isotope(70, 168),
  Yb170 = detail::isotope(70, 170),
  Yb171 = detail::isotope(70, 171),
  Yb172 = detail::isotope(70, 172),
  Yb173 = detail::isotope(70, 173),
  Yb174 = detail::isotope(70, 174),
  Yb176 = detail::isotope(70, 176),
  Lu175 = detail::isotope(71, 175),
  Lu176 = detail::isotope(71, 176),
  Hf174 = detail::isotope(72, 174),
  Hf176 = detail::isotope(72, 176),
  Hf177 = detail::isotope(72, 177),
  Hf178 = detail::isotope(72, 178),
  Hf179 = detail::isotope(72, 179),
  Hf180 = detail::isotope(72, 180),
  Ta180 = detail::isotope(73, 180),
  Ta181 = detail::isotope(73, 181),
  W180 = detail::isotope(74, 180),
  W182 = detail::isotope(74, 182),
  W183 = detail::isotope(74, 183),
  W184 = detail::isotope(74, 184),
  W186 = detail::isotope(74, 186),
  Re185 = detail::isotope(75, 185),
  Re187 = detail::isotope(75, 187),
  Os184 = detail::isotope(76, 184),
  Os186 = detail::isotope(76, 186),
  Os187 = detail::isotope(76, 187),
  Os188 = detail::isotope(76, 188),
  Os189 = detail::isotope(76, 189),
  Os190 = detail::isotope(76, 190),
  Os192 = detail::isotope(76, 192),
  Ir191 = detail::isotope(77, 191),
  Ir193 = detail::isotope(77, 193),
  Pt190 = detail::isotope(78, 190),
  Pt192 = detail::isotope(78, 192),
  Pt194 = detail::isotope(78, 194),
  Pt195 = detail::isotope(78, 195),
  Pt196 = detail::isotope(78, 196),
  Pt198 = detail::isotope(78, 198),
  Au197 = Au,
  Hg196 = detail::isotope(80, 196),
  Hg198 = detail::isotope(80, 198),
  Hg199 = detail::isotope(80, 199),
  Hg200 = detail::isotope(80, 200),
  Hg201 = detail::isotope(80, 201),
  Hg202 = detail::isotope(80, 202),
  Hg204 = detail::isotope(80, 204),
  Tl203 = detail::isotope(81, 203),
  Tl205 = detail::isotope(81, 205),
  Pb204 = detail::isotope(82, 204),
  Pb206 = detail::isotope(82, 206),
  Pb207 = detail::isotope(82, 207),
  Pb208 = detail::isotope(82, 208),
  Bi209 = Bi,
  Po209 = detail::isotope(84, 209),
  Po210 = detail::isotope(84, 210),
  At210 = detail::isotope(85, 210),
  At211 = detail::isotope(85, 211),
  Rn211 = detail::isotope(86, 211),
  Rn220 = detail::isotope(86, 220),
  Rn222 = detail::isotope(86, 222),
  Fr223 = Fr,
  Ra223 = detail::isotope(88, 223),
  Ra224 = detail::isotope(88, 224),
  Ra226 = detail::isotope(88, 226),
  Ra228 = detail::isotope(88, 228),
  Ac227 = Ac,
  Th230 = detail::isotope(90, 230),
  Th232 = detail::isotope(90, 232),
  Pa231 = Pa,
  U233 = detail::isotope(92, 233),
  U234 = detail::isotope(92, 234),
  U235 = detail::isotope(92, 235),
  U236 = detail::isotope(92, 236),
  U238 = detail::isotope(92, 238),
  Np236 = detail::isotope(93, 236),
  Np237 = detail::isotope(93, 237),
  Pu238 = detail::isotope(94, 238),
  Pu239 = detail::isotope(94, 239),
  Pu240 = detail::isotope(94, 240),
  Pu241 = detail::isotope(94, 241),
  Pu242 = detail::isotope(94, 242),
  Pu244 = detail::isotope(94, 244),
  Am241 = detail::isotope(95, 241),
  Am243 = detail::isotope(95, 243),
  Cm243 = detail::isotope(96, 243),
  Cm244 = detail::isotope(96, 244),
  Cm245 = detail::isotope(96, 245),
  Cm246 = detail::isotope(96, 246),
  Cm247 = detail::isotope(96, 247),
  Cm248 = detail::isotope(96, 248),
  Bk247 = detail::isotope(97, 247),
  Bk249 = detail::isotope(97, 249),
  Cf249 = detail::isotope(98, 249),
  Cf250 = detail::isotope(98, 250),
  Cf251 = detail::isotope(98, 251),
  Cf252 = detail::isotope(98, 252),
  Es252 = Es,
  Fm257 = Fm,
  Md258 = detail::isotope(101, 258),
  Md260 = detail::isotope(101, 260),
  No259 = No,
  Lr262 = Lr,
  Rf267 = Rf,
  Db268 = Db,
  Sg271 = Sg,
  Bh272 = Bh,
  Hs270 = Hs,
  Mt276 = Mt,
  Ds281 = Ds,
  Rg280 = Rg,
  Cn285 = Cn
};
// clang-format on

} /* namespace Utils */
} /* namespace Scine */

#endif // UTILS_ELEMENTTYPES_H_
